#! /bin/sh
exec guile-cdplayer -s "$0" "$@"
!#
;;     This file is part of cdloop
;;     Copyright (C) 2001-2003 Claus Brunzema <mail@cbrunzema.de>
;; 
;;     This program is free software; you can redistribute it and/or modify
;;     it under the terms of the GNU General Public License as published by
;;     the Free Software Foundation; either version 2 of the License, or
;;     (at your option) any later version.
;; 
;;     This program is distributed in the hope that it will be useful,
;;     but WITHOUT ANY WARRANTY; without even the implied warranty of
;;     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;     GNU General Public License for more details.
;; 
;;     You should have received a copy of the GNU General Public License
;;     along with this program; if not, write to the Free Software
;;     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;; 

(debug-enable 'backtrace)
(debug-enable 'trace)
(debug-enable 'procnames)

(use-modules (ice-9 format))
(use-modules (ice-9 getopt-long))

;;;{{{ defaults

(define cdloop-config-file "~/.cdlooprc")
(define cdloop-bookmark-file "~/.cdloop-loops")
(define cdloop-device "/dev/cdrom")
(define cdloop-gui-update-interval-milliseconds 100)
(define cdloop-mixer-enabled #t)
(define start-volume-right 1.0)
(define start-volume-left 1.0)
(define start-mute-left #f)
(define start-mute-right #f)
(define search-frames-backward 150)
(define search-frames-forward 150)
(define toplevel-width 575)
(define toplevel-height 580)
(define autoset-locatorX #t)
(define autoset-locatorB-offset-frames (* 5 75))
(define autoset-gap-microseconds 1000000)

;;;}}}
;;;{{{ misc util

(define (expand-filename filename)
  (if (char=? (string-ref filename 0) #\~)
      (string-append (getenv "HOME") (substring filename 1))
      filename))

(define pos-track car)
(define pos-minute cadr)
(define pos-second caddr)
(define pos-frame cadddr)

(define (pos->string pos)
  (if (and (pair? pos) (> (pos-track pos) 0))
      (format "~2,'0d-~2,'0d:~2,'0d:~2,'0d"
              (pos-track pos)
              (pos-minute pos)
              (pos-second pos)
              (pos-frame pos))
      "-----------"))

(define (status->string status)
  (cond ((not status) "INIT")
        (else (substring (symbol->string status) 7)))) ; erase STATUS_

(define (do-quit . dummy)
  (if timeout-handle (gtk-timeout-remove timeout-handle))
  (save-bookmarks)
  (cdplayer-exit)
  (gtk-exit))

;;;}}}
;;;{{{ bookmark data type

(define bookmark-rtd
  (make-record-type "Bookmark"
                    (list 'id
                          'name
                          'locatorA
                          'locatorB
                          'gap
                          'volumeL
                          'volumeR
                          'muteL
                          'muteR)))
(define make-bookmark
  (record-constructor bookmark-rtd))
(define bookmark-get-id
  (record-accessor bookmark-rtd 'id))
(define bookmark-set-id
  (record-modifier bookmark-rtd 'id))
(define bookmark-get-name
  (record-accessor bookmark-rtd 'name))
(define bookmark-set-name
  (record-modifier bookmark-rtd 'name))
(define bookmark-get-locatorA
  (record-accessor bookmark-rtd 'locatorA))
(define bookmark-set-locatorA
  (record-modifier bookmark-rtd 'locatorA))
(define bookmark-get-locatorB
  (record-accessor bookmark-rtd 'locatorB))
(define bookmark-set-locatorB
  (record-modifier bookmark-rtd 'locatorB))
(define bookmark-get-gap
  (record-accessor bookmark-rtd 'gap))
(define bookmark-set-gap
  (record-modifier bookmark-rtd 'gap))
(define bookmark-get-volumeL
  (record-accessor bookmark-rtd 'volumeL))
(define bookmark-set-volumeL
  (record-modifier bookmark-rtd 'volumeL))
(define bookmark-get-volumeR
  (record-accessor bookmark-rtd 'volumeR))
(define bookmark-set-volumeR
  (record-modifier bookmark-rtd 'volumeR))
(define bookmark-get-muteL
  (record-accessor bookmark-rtd 'muteL))
(define bookmark-set-muteL
  (record-modifier bookmark-rtd 'muteL))
(define bookmark-get-muteR
  (record-accessor bookmark-rtd 'muteR))
(define bookmark-set-muteR
  (record-modifier bookmark-rtd 'muteR))
(define (bookmark-display bm)
  (display (format "(make-bookmark ~a \"~a\" (quote ~a) (quote ~a) ~a ~a ~a ~a ~a)"
                   (bookmark-get-id bm)
                   (bookmark-get-name bm)
                   (bookmark-get-locatorA bm)
                   (bookmark-get-locatorB bm)
                   (bookmark-get-gap bm)
                   (bookmark-get-volumeL bm)
                   (bookmark-get-volumeR bm)
                   (bookmark-get-muteL bm)
                   (bookmark-get-muteR bm))))

;;;}}}
;;;{{{ globals

(define cdloop-version "CDLOOP-VERSION")
(define cdloop-about-text (string-append
"This is cdloop " cdloop-version "\n"
"(C) 2001-2003 Claus Brunzema <mail@cbrunzema.de>\n"
"see http://www.cbrunzema.de/software.html#cdloop\n"
"released under the GNU General Public License Version 2\n"
))
(define cdloop-help-text (string-append
;"Keys:\n"
;"     q        quit\n"
;"     0-9      play track (0 -> play track 10)\n"
;"     p        pause/resume\n"
;"     a        set locator A\n"
;"     b        set locator B\n"
;"     x        set locator X\n"
;"\n"
"Options:\n"
"        -h, --help                     show help message\n"
"        -v, --version                  show version info\n"
"        -l, --load=<file>              load alternate configuration file\n"
"                                       (default: " cdloop-config-file ")\n"
"        -e, --eval=<form>              eval the form after loading the\n"
"                                       configuration file. Quote carefully.\n"
"        -d, --device=<device>          use this special device as cd drive\n"
"                                       (default: " cdloop-device ")\n"
"        --no-mixer                     don't use the mixer\n"
"        -b, --bookmark=<bookmarkfile>  use this bookmark file\n"
"                                       (default: " cdloop-bookmark-file ")\n"
))
(define timeout-handle #f)
(define name "")
(define locatorA #f)
(define locatorB #f)
(define locatorX #f)
(define gap 0.0)
(define gap-learn-mode #f)
(define volume-right 1.0)
(define volume-left 1.0)
(define mute-left #f)
(define mute-right #f)
(define disc-id #f)
(define bookmarks '())
(define current-bookmark #f)
(define loop-start #f)

;;;}}}
;;;{{{ hooks
(define name-changed-hook (make-hook 1))
(define locatorA-changed-hook (make-hook 1))
(define locatorB-changed-hook (make-hook 1))
(define locatorX-changed-hook (make-hook 1))
(define gap-changed-hook (make-hook 1))
(define gap-learn-mode-changed-hook (make-hook 1))
(define volumeL-changed-hook (make-hook 1))
(define volumeR-changed-hook (make-hook 1))
(define muteL-changed-hook (make-hook 1))
(define muteR-changed-hook (make-hook 1))
(define status-changed-hook (make-hook 1))
(define pos-changed-hook (make-hook 1))
(define toc-changed-hook (make-hook))
(define bookmarks-changed-hook (make-hook))
(define current-bookmark-changed-hook (make-hook 1))
(define loop-status-changed-hook (make-hook 1))
(define loop-start-changed-hook (make-hook 1))
;;;}}}
;;;{{{ bookmark functions

(define (save-bookmarks)
  (with-output-to-file cdloop-bookmark-file
    (lambda ()
      (display "(set! bookmarks (list")(newline)
      (for-each (lambda (bm)
                  (bookmark-display bm)
                  (newline))
                bookmarks)
      (display "))")(newline))))

(define (set-current-bookmark index)
  (set! current-bookmark index)
  (if index
      (let ((bm (list-ref bookmarks index)))
        (set-name (bookmark-get-name bm))
        (set-locatorA (bookmark-get-locatorA bm))
        (set-loop-start 'A)
        (set-locatorB (bookmark-get-locatorB bm))
        (set-gap (bookmark-get-gap bm))
        (set-volumeL (bookmark-get-volumeL bm))
        (set-volumeR (bookmark-get-volumeR bm))
        (set-muteL (bookmark-get-muteL bm))
        (set-muteR (bookmark-get-muteR bm))
        (if (cdplayer-pos)
            (cdplayer-play locatorA)))
      (set-name ""))
  (run-hook current-bookmark-changed-hook index))

(define (register-bookmark)
  (set! bookmarks
        (append bookmarks
                (list (make-bookmark (cdplayer-id)
                                     name
                                     locatorA
                                     locatorB
                                     gap
                                     volume-left volume-right
                                     mute-left mute-right))))
  (run-hook bookmarks-changed-hook))

(define (change-current-bookmark)
  (if current-bookmark
      (let ((bm (list-ref bookmarks current-bookmark)))
        (bookmark-set-name bm name)
        (bookmark-set-locatorA bm locatorA)
        (bookmark-set-locatorB bm locatorB)
        (bookmark-set-gap bm gap)
        (bookmark-set-volumeL bm volume-left)
        (bookmark-set-volumeR bm volume-right)
        (bookmark-set-muteL bm mute-left)
        (bookmark-set-muteR bm mute-right)
        (run-hook bookmarks-changed-hook))))

(define (delete-current-bookmark)
  (cond (current-bookmark
         (set! bookmarks (delete (list-ref bookmarks current-bookmark)
                                 bookmarks))
         (set-current-bookmark #f)
         (run-hook bookmarks-changed-hook))))

;;;}}}
;;;{{{ global set functions
(define (set-name n)
  (set! name n)
  (run-hook name-changed-hook name))
(define (set-locatorA pos)
  (set! locatorA pos)
  (run-hook locatorA-changed-hook locatorA))
(define (set-locatorB pos)
  (set! locatorB pos)
  (run-hook locatorB-changed-hook locatorB))
(define (set-locatorX pos)
  (set! locatorX pos)
  (run-hook locatorX-changed-hook locatorX))
(define (set-gap g)
  (set! gap g)
  (run-hook gap-changed-hook gap))
(define (set-gap-learn-mode bool)
  (set! gap-learn-mode bool)
  (set-gap 100000000)
  (run-hook gap-learn-mode-changed-hook bool))
(define (set-loop-start loop)
  (set! loop-start loop)
  (run-hook loop-start-changed-hook loop))
;;;}}}
;;;{{{ mixer functions
(define (set-volumeL vol)
  (if cdloop-mixer-enabled
      (begin
        (set! volume-left vol)
        (cdplayer-mixer-set-master-volume-front-left vol)
        (run-hook volumeL-changed-hook vol))))
(define (set-volumeR vol)
  (if cdloop-mixer-enabled
      (begin
        (set! volume-right vol)
        (cdplayer-mixer-set-master-volume-front-right vol)
        (run-hook volumeR-changed-hook vol))))
(define (set-muteL mute)
  (if cdloop-mixer-enabled
      (begin
        (set! mute-left mute)
        (cdplayer-mixer-set-mute-front-left mute)
        (run-hook muteL-changed-hook mute))))
(define (set-muteR mute)
  (if cdloop-mixer-enabled
      (begin
        (set! mute-right mute)
        (cdplayer-mixer-set-mute-front-right mute)
        (run-hook muteR-changed-hook mute))))
(define (reset-mixer)
  (if cdloop-mixer-enabled
      (begin
        (set-volumeL start-volume-left)
        (set-volumeR start-volume-right)
        (set-muteL start-mute-left)
        (set-muteR start-mute-right))))
;;;}}}
;;;{{{ cd util

(define (play-track track)
  (cdplayer-play (list track 0 0 0)))

(define (calc-pos-and-do pos frameoffset func)
  (let ((new-pos (cdplayer-add-pos pos (list 0 0 frameoffset))))
    (if new-pos
        (func new-pos))))

(define (get-current-locator)
  (cond
   ((eq? 'A loop-start) locatorA)
   ((eq? 'X loop-start) locatorX)))

(define (set-loop-if-valid dummy)
  (if (and (get-current-locator) locatorB) 
      (begin
        (cdplayer-set-loop (list (get-current-locator) locatorB gap))
        (run-hook loop-status-changed-hook #t))
      (begin
        (cdplayer-clear-loop)
        (run-hook loop-status-changed-hook #f))))

(define (do-play)
  (if (eq? 'STATUS_PAUSE (cdplayer-status))
      (cdplayer-resume)
      (if (get-current-locator)
          (cdplayer-play (get-current-locator))
          (let ((pos (cdplayer-pos)))
            (if pos
                (play-track (pos-track pos))
                (play-track 1))))))

(define (do-pause)
  (let ((status (cdplayer-status)))
    (cond ((eq? 'STATUS_PAUSE status)
           (cdplayer-resume))
          ((or (eq? 'STATUS_PLAY status)
               (eq? 'STATUS_LOOP status))
           (cdplayer-pause)))))

(define (do-set-locatorA)
  (set-locatorA (cdplayer-pos)))

(define (do-set-locatorB)
  (let ((pos (cdplayer-pos)))
    (if autoset-locatorX
        (begin
          (calc-pos-and-do pos (- autoset-locatorB-offset-frames)
                           set-locatorX)
          (set-gap autoset-gap-microseconds)
          (set-loop-start 'X)))
    (set-locatorB pos)))

(define (do-set-locatorX)
  (set-locatorX (cdplayer-pos)))

(define (do-clear-loop)
  (set-locatorA #f)
  (set-locatorB #f)
  (set-locatorX #f)
  (set-gap 0)
  (set-current-bookmark #f))

;;;}}}
;;;{{{ gui util
(define gui-gap-display-factor 1000000)
(define gui-box-spacing 5)

(define (make-button text callback)
  (let ((button (gtk-button-new-with-label text)))
    (gtk-signal-connect button "clicked" callback)
    (gtk-widget-show button)
    button))

(define (show-about)
  (let ((win (gtk-window-new 'toplevel))
        (vbox (gtk-vbox-new #f 0))
        (label (gtk-label-new cdloop-about-text)))
    (define (close-about . dummy)
      (gtk-widget-destroy win))
    (gtk-window-set-title win "About cdloop")
    (gtk-window-set-policy win #t #t #t)
    (gtk-signal-connect win "delete-event" close-about)
    (gtk-signal-connect win "destroy-event" close-about)
    (gtk-label-set-justify label 'left)
    (gtk-container-border-width win gui-box-spacing)
    (gtk-container-add win vbox)
    (gtk-box-pack-start vbox label)
    (gtk-box-pack-start vbox (make-button "Close" close-about))
    (gtk-widget-show-all win)))
;;;}}}
;;;{{{ gui
(define (make-gui)
;;;{{{ status and command group

  (define (make-status-group)
    (let ((hbox (gtk-hbox-new #t 0))
          (status-label (gtk-label-new (status->string #f)))
          (pos-label (gtk-label-new (pos->string #f))))
      (gtk-box-pack-start hbox status-label)
      (gtk-box-pack-start hbox pos-label)
      (add-hook!
       status-changed-hook
         (lambda (status)
                  (gtk-label-set-text status-label 
                                      (status->string status))))
      (add-hook!
       pos-changed-hook
         (lambda (pos)
                  (gtk-label-set-text pos-label (pos->string pos))))
      hbox))

  (define (make-command-group)
    (let ((frame (gtk-frame-new "Control"))
          (vbox (gtk-vbox-new #f 0)))
      (gtk-container-add frame vbox)
      (add-hook! gap-learn-mode-changed-hook
                 (lambda (state)
                   (gtk-widget-set-sensitive frame (not state))))
      (gtk-box-pack-start
       vbox
       (let ((hbox (gtk-hbox-new #t 0)))
         (gtk-box-pack-start hbox (make-button "Play" do-play))
         (gtk-box-pack-start hbox (make-button "Pause" do-pause))
         (gtk-box-pack-start hbox (make-button "Stop" cdplayer-stop))
         hbox))
      (gtk-box-pack-start
       vbox
       (let ((hbox (gtk-hbox-new #t 0)))
         (gtk-box-pack-start hbox (make-button
                                   "|< Skip"
                                   (lambda ()
                                     (let ((pos (cdplayer-pos)))
                                       (if pos
                                           (if (< (pos-second pos) 1)
                                               (play-track (- (pos-track pos) 1))
                                               (play-track (pos-track pos))))))))
         (gtk-box-pack-start hbox (make-button
                                   "Skip >|"
                                   (lambda ()
                                     (let ((pos (cdplayer-pos)))
                                       (if pos
                                           (play-track (+ (pos-track pos) 1)))))))
         (gtk-box-pack-start hbox (make-button
                                   "< Search"
                                   (lambda ()
                                     (calc-pos-and-do
                                      (cdplayer-pos) (- search-frames-backward)
                                      cdplayer-play))))
         (gtk-box-pack-start hbox (make-button
                                   "Search >"
                                   (lambda ()
                                     (calc-pos-and-do
                                      (cdplayer-pos) search-frames-forward
                                      cdplayer-play))))
         hbox))

      (gtk-box-pack-start
       vbox
       (let ((hbox (gtk-hbox-new #t 0)))
         (gtk-box-pack-start hbox (make-button "Eject" cdplayer-eject))
         (gtk-box-pack-start hbox (make-button "Close" cdplayer-close-tray))
         hbox))
      frame))

;;;}}}
;;;{{{ track table

  (define (make-track-table)
    (let ((trackframe (gtk-frame-new "Tracks"))
          (trackbutton-table (gtk-table-new 10 10 #t))
          (trackbutton-list '()))
      (gtk-container-add trackframe trackbutton-table)
      (add-hook! gap-learn-mode-changed-hook
                 (lambda (state)
                   (gtk-widget-set-sensitive trackframe (not state))))
      (let ((track 1))
        (do ((row 0 (1+ row)))
            ((or (> track 99) (>= row 10)))
          (do ((col 0 (1+ col)))
              ((or (> track 99) (>= col 10)))
            (let ((button (make-button (number->string track)
                                       (let ((button-track track))
                                         (lambda ()
                                           (play-track button-track))))))
              (set! trackbutton-list (append trackbutton-list (list button)))
              (gtk-table-attach-defaults
               trackbutton-table button
               col (+ 1 col) row (+ 1 row))
              (set! track (+ 1 track))))))
      (add-hook! toc-changed-hook
                 (lambda ()
                   (let ((count (cdplayer-no-of-tracks)))
                     (do ((track 0 (+ 1 track)))
                         ((> track 98))
                       (gtk-widget-set-sensitive
                        (list-ref trackbutton-list track)
                        (< track count))))))
      trackframe))

;;;}}}
;;;{{{ mixer group

  (define (make-mixer-group)
     (let ((frame (gtk-frame-new "Mixer"))
           (vbox (gtk-vbox-new #f 0)))
       (gtk-container-add frame vbox)
       (gtk-box-pack-start
        vbox
        (let ((hbox (gtk-hbox-new #t 0)))
          (gtk-box-pack-start
           hbox
           (let* ((vbox (gtk-vbox-new #f 0))
                  (adjustment (gtk-adjustment-new
                               (- 1.0 volume-left)
                               0.0 1.0
                               0.1
                               0.1 0.1))
                  (scrollbar (gtk-vscrollbar-new adjustment))
                  (mute-checkbutton (gtk-check-button-new)))
             (gtk-range-set-update-policy scrollbar 'continuous)
             (gtk-box-pack-start vbox (gtk-label-new "L") #f #f)
             (gtk-box-pack-start vbox scrollbar)
             (gtk-signal-connect
              adjustment
              "value_changed"
              (lambda ()
                (set-volumeL (- 1.0 (gtk-adjustment-value adjustment)))))
             (add-hook! volumeL-changed-hook
                        (lambda (v)
                          (if (not (= v (gtk-adjustment-value adjustment)))
                              (gtk-adjustment-set-value
                               adjustment (- 1.0 v)))))
             (gtk-box-pack-start vbox mute-checkbutton #f #f)
             (gtk-signal-connect
              mute-checkbutton
              "toggled"
              (lambda ()
                (set-muteL (gtk-toggle-button-active mute-checkbutton))))
             (add-hook! muteL-changed-hook
                        (lambda (m)
                          (if (not (eq? m (gtk-toggle-button-active
                                           mute-checkbutton)))
                              (gtk-toggle-button-set-state
                               mute-checkbutton m))))
             vbox) #f #f)
          (gtk-box-pack-start
           hbox
           (let* ((vbox (gtk-vbox-new #f 0))
                  (adjustment (gtk-adjustment-new
                               (- 1.0 volume-right)
                               0.0 1.0
                               0.1
                               0.1 0.1))
                  (scrollbar (gtk-vscrollbar-new adjustment))
                  (mute-checkbutton (gtk-check-button-new)))
             (gtk-range-set-update-policy scrollbar 'continuous)
             (gtk-box-pack-start vbox (gtk-label-new "R") #f #f)
             (gtk-box-pack-start vbox scrollbar)
             (gtk-signal-connect
              adjustment
              "value_changed"
              (lambda ()
                (set-volumeR (- 1.0 (gtk-adjustment-value adjustment)))))
             (add-hook! volumeR-changed-hook
                        (lambda (v)
                          (if (not (= v (gtk-adjustment-value adjustment)))
                              (gtk-adjustment-set-value
                               adjustment (- 1.0 v)))))
             (gtk-box-pack-start vbox mute-checkbutton #f #f)
             (gtk-signal-connect
              mute-checkbutton
              "toggled"
              (lambda ()
                (set-muteR (gtk-toggle-button-active mute-checkbutton))))
             (add-hook! muteR-changed-hook
                        (lambda (m)
                          (if (not (eq? m (gtk-toggle-button-active
                                           mute-checkbutton)))
                              (gtk-toggle-button-set-state
                               mute-checkbutton m))))
             vbox) #f #f)
          hbox) #t #t)
       (gtk-box-pack-start vbox (make-button "Reset" reset-mixer) #f #f)
       frame))

;;;}}}
;;;{{{ locator groups and gap group
;;;{{{ locator A

  (define (make-locatorA-group loop-start-radio-button-group)
    (let ((frame (gtk-frame-new "Locator A"))
          (vbox (gtk-vbox-new #f 0))
          (label (gtk-label-new (pos->string #f)))
          (loop-start-button (gtk-radio-button-new-with-label-from-widget
                              loop-start-radio-button-group
                              "Start loop with locator A")))
      (add-hook!
       locatorA-changed-hook
       (lambda (locator)
         (gtk-label-set-text label (pos->string locator))))
      (gtk-container-add frame vbox)
      (add-hook! gap-learn-mode-changed-hook
                 (lambda (state)
                   (gtk-widget-set-sensitive frame (not state))))
      
      (gtk-box-pack-start
       vbox
       (let ((hbox (gtk-hbox-new #t 0)))
         (gtk-box-pack-start hbox (make-button
                                   "Jump"
                                   (lambda ()
                                     (if locatorA
                                         (cdplayer-play locatorA)))) #t #t)
         (gtk-box-pack-start hbox (gtk-label-new " ") #t #t)
         hbox) #f #f)
      
      (gtk-box-pack-start
       vbox
       (let ((hbox (gtk-hbox-new #t 0)))
         (gtk-box-pack-start hbox (make-button "Set" do-set-locatorA))
         (gtk-box-pack-start hbox (make-button
                                   "Clear"
                                   (lambda ()
                                     (set-locatorA #f))))
         hbox))
      
      (gtk-box-pack-start
       vbox
       (let ((hbox (gtk-hbox-new #f 0)))
         (gtk-box-pack-start
          hbox
          (let ((hbox (gtk-hbox-new #t 0)))
            (gtk-box-pack-start hbox (make-button
                                      "<<<"
                                      (lambda ()
                                        (calc-pos-and-do locatorA
                                                         -50
                                                         set-locatorA))))
            (gtk-box-pack-start hbox (make-button
                                      "<<"
                                      (lambda ()
                                        (calc-pos-and-do locatorA
                                                         -10
                                                         set-locatorA))))
            (gtk-box-pack-start hbox (make-button
                                      "<"
                                      (lambda ()
                                        (calc-pos-and-do locatorA
                                                         -1
                                                         set-locatorA))))
            hbox))
         (gtk-box-pack-start hbox label)
         (gtk-box-pack-start
          hbox
          (let ((hbox (gtk-hbox-new #t 0)))
            (gtk-box-pack-start hbox (make-button
                                      ">"
                                      (lambda ()
                                        (calc-pos-and-do locatorA
                                                         1
                                                         set-locatorA))))
            (gtk-box-pack-start hbox (make-button
                                      ">>"
                                      (lambda ()
                                        (calc-pos-and-do locatorA
                                                         10
                                                         set-locatorA))))
            (gtk-box-pack-start hbox (make-button
                                      ">>>"
                                      (lambda ()
                                        (calc-pos-and-do locatorA
                                                         50
                                                         set-locatorA))))
            hbox))
         hbox))
      
      (gtk-box-pack-start vbox loop-start-button)
      (gtk-signal-connect
       loop-start-button 
       "toggled"
       (lambda ()
         (if (gtk-toggle-button-active loop-start-button)
             (set-loop-start 'A))))
      (add-hook! loop-start-changed-hook
                 (lambda (start)
                   (gtk-toggle-button-set-state loop-start-button
                                                (eq? start 'A))))
      (list frame loop-start-button)))

;;;}}}
;;;{{{ locator X
  (define (make-locatorX-group loop-start-radio-button-group)
    (let ((frame (gtk-frame-new "Locator X"))
          (vbox (gtk-vbox-new #f 0))
          (label (gtk-label-new (pos->string #f)))
          (loop-start-button (gtk-radio-button-new-with-label-from-widget
                              loop-start-radio-button-group
                              "Start loop with locator X")))
      (add-hook!
       locatorX-changed-hook
       (lambda (locator)
         (gtk-label-set-text label (pos->string locator))))
      (gtk-container-add frame vbox)
      (add-hook! gap-learn-mode-changed-hook
                 (lambda (state)
                   (gtk-widget-set-sensitive frame (not state))))
      (gtk-box-pack-start
       vbox
       (let ((hbox (gtk-hbox-new #t 0)))
         (gtk-box-pack-start hbox (make-button
                                   "Jump"
                                   (lambda ()
                                     (if locatorX
                                         (cdplayer-play locatorX)))) #t #t)
         (gtk-box-pack-start hbox (gtk-label-new " ") #t #t)
         hbox) #f #f)
         
      (gtk-box-pack-start
       vbox
       (let ((hbox (gtk-hbox-new #t 0)))
         (gtk-box-pack-start hbox (make-button "Set" do-set-locatorX))
         (gtk-box-pack-start hbox (make-button
                                   "Clear"
                                   (lambda ()
                                     (set-locatorX #f))))
         hbox))
      
      (gtk-box-pack-start
       vbox
       (let ((hbox (gtk-hbox-new #f 0)))
         (gtk-box-pack-start
          hbox
          (let ((hbox (gtk-hbox-new #t 0)))
            (gtk-box-pack-start hbox (make-button
                                      "<<<"
                                      (lambda ()
                                        (calc-pos-and-do locatorX
                                                         -50
                                                         set-locatorX))))
            (gtk-box-pack-start hbox (make-button
                                      "<<"
                                      (lambda ()
                                        (calc-pos-and-do locatorX
                                                         -10
                                                         set-locatorX))))
            (gtk-box-pack-start hbox (make-button
                                      "<"
                                      (lambda ()
                                        (calc-pos-and-do locatorX
                                                         -1
                                                         set-locatorX))))
            hbox))
         (gtk-box-pack-start hbox label)
         (gtk-box-pack-start
          hbox
          (let ((hbox (gtk-hbox-new #t 0)))
            (gtk-box-pack-start hbox (make-button
                                      ">"
                                      (lambda ()
                                        (calc-pos-and-do locatorX
                                                         1
                                                         set-locatorX))))
            (gtk-box-pack-start hbox (make-button
                                      ">>"
                                      (lambda ()
                                        (calc-pos-and-do locatorX
                                                         10
                                                         set-locatorX))))
            (gtk-box-pack-start hbox (make-button
                                      ">>>"
                                      (lambda ()
                                        (calc-pos-and-do locatorX
                                                         50
                                                         set-locatorX))))
            hbox))
         hbox))
      
      (gtk-box-pack-start vbox loop-start-button)
      (gtk-signal-connect
       loop-start-button 
       "toggled"
       (lambda ()
         (if (gtk-toggle-button-active loop-start-button)
             (set-loop-start 'X))))
      (add-hook! loop-start-changed-hook
                 (lambda (start)
                   (gtk-toggle-button-set-state loop-start-button
                                                (eq? start 'X))))
      (list frame loop-start-button)))
;;;}}}
;;;{{{ locator B
  (define (make-locatorB-group)
    (let ((frame (gtk-frame-new "Locator B"))
          (vbox (gtk-vbox-new #f 0))
          (label (gtk-label-new (pos->string #f))))
      (add-hook!
       locatorB-changed-hook
       (lambda (locator)
         (gtk-label-set-text label (pos->string locator))))
      (gtk-container-add frame vbox)
      (add-hook! gap-learn-mode-changed-hook
                 (lambda (state)
                   (gtk-widget-set-sensitive frame (not state))))

      (gtk-box-pack-start
       vbox
       (let ((hbox (gtk-hbox-new #t 0)))
         (gtk-box-pack-start hbox (make-button "Set" do-set-locatorB))
         (gtk-box-pack-start hbox (make-button
                                   "Clear"
                                   (lambda ()
                                     (set-locatorB #f))))
         hbox))
      
      (gtk-box-pack-start
       vbox
       (let ((hbox (gtk-hbox-new #f 0)))
         (gtk-box-pack-start
          hbox
          (let ((hbox (gtk-hbox-new #t 0)))
            (gtk-box-pack-start hbox (make-button
                                      "<<<"
                                      (lambda ()
                                        (calc-pos-and-do locatorB
                                                         -50
                                                         set-locatorB))))
            (gtk-box-pack-start hbox (make-button
                                      "<<"
                                      (lambda ()
                                        (calc-pos-and-do locatorB
                                                         -10
                                                         set-locatorB))))
            (gtk-box-pack-start hbox (make-button
                                      "<"
                                      (lambda ()
                                        (calc-pos-and-do locatorB
                                                         -1
                                                         set-locatorB))))
            hbox))
      
         (gtk-box-pack-start hbox label)

         (gtk-box-pack-start
          hbox
          (let ((hbox (gtk-hbox-new #t 0)))
            (gtk-box-pack-start hbox (make-button
                                      ">"
                                      (lambda ()
                                        (calc-pos-and-do locatorB
                                                         1
                                                         set-locatorB))))
            (gtk-box-pack-start hbox (make-button
                                      ">>"
                                      (lambda ()
                                        (calc-pos-and-do locatorB
                                                         10
                                                         set-locatorB))))
            (gtk-box-pack-start hbox (make-button
                                      ">>>"
                                      (lambda ()
                                        (calc-pos-and-do locatorB
                                                         50
                                                         set-locatorB))))
            hbox))
         hbox))
      frame))
;;;}}}
;;;{{{ gap group

  (define (make-gap-group)
    (let* ((frame (gtk-frame-new "Gap"))
           (hbox (gtk-hbox-new #f 0))
           (adj (gtk-adjustment-new 0.0 0.0 3000.0 0.001 0.1 0.1))
           (spinbutton (gtk-spin-button-new adj 0.01 3))
           (clear-button (make-button "Clear" (lambda ()
                                                (set-gap 0))))
           (learn-button (gtk-toggle-button-new-with-label "Learn Gap")))
      (gtk-container-add frame hbox)
      (gtk-box-pack-start hbox spinbutton)
      (gtk-widget-unset-flags spinbutton '(can-focus))

      (gtk-signal-connect
       adj
       "value_changed"
       (lambda ()
         (set-gap
          (inexact->exact
           (floor (* gui-gap-display-factor (gtk-adjustment-value adj)))))))
      (add-hook!
       gap-changed-hook
       (lambda (g)
         (if gap-learn-mode
             (begin
               (gtk-widget-set-sensitive spinbutton #f)
               (gtk-widget-set-sensitive clear-button #f))
             (begin
               (gtk-widget-set-sensitive spinbutton #t)
               (gtk-widget-set-sensitive clear-button #t)
               (if (not
                    (= g (* gui-gap-display-factor
                            (gtk-adjustment-value adj))))
                   (gtk-adjustment-set-value
                    adj
                    (/ g gui-gap-display-factor)))))))
      (gtk-box-pack-start hbox clear-button)
      (gtk-widget-set-sensitive learn-button #f)
      (gtk-box-pack-start hbox learn-button)
      (add-hook!
       loop-status-changed-hook
       (lambda (status)
         (gtk-widget-set-sensitive learn-button status)))
      (gtk-signal-connect
       learn-button
       "toggled"
       (lambda ()
         (if (gtk-toggle-button-active learn-button)
             (set-gap-learn-mode #t)
             (begin
               (set-gap-learn-mode #f)
               (set-gap (cdplayer-elapsed-gap))))))
      frame))

;;;}}}
;;;}}}
;;;{{{ bookmark group

  (define (make-bookmark-group)
    (let* ((frame (gtk-frame-new "Bookmarks"))
           (vbox (gtk-vbox-new #f gui-box-spacing))
           (titles #("Name" " Locator A " " Locator B " " Gap "
                     " Vol L " " Vol R "
                     " Mute L " " Mute R "))
           (bookmarks-clist (gtk-clist-new-with-titles titles))
           (scrolled-win (gtk-scrolled-window-new))
           (bookmarks-clist-row-to-bookmark-index '()))

      (define (update-bookmark-clist)
        (gtk-clist-freeze bookmarks-clist)
        (gtk-clist-clear bookmarks-clist)
        (set! bookmarks-clist-row-to-bookmark-index '())
        (let ((index 0)
              (id (cdplayer-id)))
          (if id
              (for-each
               (lambda (bm)
                 (cond
                  ((= id (bookmark-get-id bm))
                   (gtk-clist-append
                    bookmarks-clist
                    (vector (bookmark-get-name bm)
                            (pos->string (bookmark-get-locatorA bm))
                            (pos->string (bookmark-get-locatorB bm))
                            (format "~,3f" (/ (bookmark-get-gap bm)
                                              gui-gap-display-factor))
                            (format
                             "~3d"
                             (inexact->exact
                              (* 100 (bookmark-get-volumeL bm))))
                            (format
                             "~3d"
                             (inexact->exact
                              (* 100 (bookmark-get-volumeR bm))))
                            (if (bookmark-get-muteL bm) "Y" "N")
                            (if (bookmark-get-muteR bm) "Y" "N")))
                   (set! bookmarks-clist-row-to-bookmark-index
                         (append
                          bookmarks-clist-row-to-bookmark-index
                          (list index)))))
                 (set! index (+ 1 index)))
              bookmarks)))
        (gtk-clist-thaw bookmarks-clist))
      
      (gtk-container-add frame vbox)
      (add-hook! gap-learn-mode-changed-hook
                 (lambda (state)
                   (gtk-widget-set-sensitive frame (not state))))
      (gtk-box-pack-start
       vbox
       (let ((hbox (gtk-hbox-new #f gui-box-spacing))
             (label (gtk-label-new "Name:"))
             (name-entry (gtk-entry-new)))
         (gtk-box-pack-start hbox label #f #f)
         (gtk-box-pack-start hbox name-entry)
         (gtk-signal-connect name-entry "changed"
                             (lambda ()
                               (set-name (gtk-entry-get-text name-entry))))
         (add-hook! name-changed-hook
                    (lambda (n)
                      (if (not (string=? n (gtk-entry-get-text name-entry)))
                          (gtk-entry-set-text name-entry n))))
         (gtk-box-pack-start
          hbox
          (make-button "Add Bookmark" register-bookmark))
         (gtk-box-pack-start
          hbox
          (make-button "Change Bookmark" change-current-bookmark))
         (gtk-box-pack-start
          hbox
          (make-button "Delete Bookmark" delete-current-bookmark))
         hbox) #f #f)

      (gtk-scrolled-window-set-policy scrolled-win 'automatic 'automatic)
      (gtk-clist-set-selection-mode bookmarks-clist 'single)
      (gtk-clist-set-column-width bookmarks-clist 0 143);; magic constant FIXME
      (gtk-clist-set-column-auto-resize bookmarks-clist 1 #t)
      (gtk-clist-set-column-auto-resize bookmarks-clist 2 #t)
      (gtk-clist-set-column-auto-resize bookmarks-clist 3 #t)
      (gtk-clist-set-column-auto-resize bookmarks-clist 4 #t)
      (gtk-clist-set-column-auto-resize bookmarks-clist 5 #t)
      (gtk-clist-set-column-auto-resize bookmarks-clist 6 #t)
      (gtk-clist-set-column-justification bookmarks-clist 6 'center)
      (gtk-clist-set-column-auto-resize bookmarks-clist 7 #t)
      (gtk-clist-set-column-justification bookmarks-clist 7 'center)
      (gtk-signal-connect bookmarks-clist "select_row"
                          (lambda (row col event)
                            (set-current-bookmark
                             (list-ref
                              bookmarks-clist-row-to-bookmark-index
                              row))))
      (gtk-signal-connect bookmarks-clist "unselect_row"
                          (lambda (row col event)
                            (set-current-bookmark #f)))
      (gtk-container-add scrolled-win bookmarks-clist)
      (gtk-box-pack-start vbox scrolled-win #t #t)
      (add-hook! bookmarks-changed-hook update-bookmark-clist)
      (add-hook!
       current-bookmark-changed-hook
       (lambda (index)
         (if index
             (begin
               ;; FIXME: set selected index in clist
               #t)
             (let ((selection-row (gtk-clist-selection bookmarks-clist)))
               (if (pair? selection-row)
                   (gtk-clist-unselect-row bookmarks-clist
                                           (car selection-row)
                                           0))))))
      frame))

;;;}}}
;;;{{{ toplevel

  (let ((win (gtk-window-new 'toplevel))
        (vbox (gtk-vbox-new #f gui-box-spacing)))
    (gtk-window-set-title win (string-append "cdloop " cdloop-device))
    (gtk-window-set-policy win #t #t #t)
    (gtk-signal-connect win "delete-event" do-quit)
    (gtk-signal-connect win "destroy-event" do-quit)
    (gtk-widget-set-usize win toplevel-width toplevel-height)

;; help desparately needed for keyboard control. the method below
;; sucks, because e.g. it quits the app if you type a 'q' in the
;; name entry...
;    (gtk-signal-connect
;     win
;     "key_press_event"
;     (lambda (event)
;       (let ((keyval (gdk-event-keyval event)))
;;       (display keyval) (newline)
;        (cond
;         ((= keyval 113) ; q
;          (do-quit))
;         ((= keyval 101) ; e
;          (cdplayer-eject))
;         ((= keyval 99) ; c
;          (cdplayer-close-tray))
;         ((= keyval 97)  ; a
;          (do-set-locatorA))
;         ((and (= keyval 65) locatorA)  ; A
;          (cdplayer-play locatorA))
;         ((= keyval 98)  ; b
;          (do-set-locatorB))
;         ((and (= keyval 66) locatorB)  ; B
;          (cdplayer-play locatorB))
;         ((= keyval 120) ; x
;          (do-set-locatorX))
;         ((and (= keyval 88) locatorX)  ; X
;          (cdplayer-play locatorX))
;         ((= keyval 112) ; p
;          (do-pause))
;         ((= keyval 80) ; P
;          (do-play))
;         ((and (>= keyval 49) (<= keyval 58)) ; 1-9
;          (play-track (- keyval 48)))
;         ((= keyval 48)  ; 0
;          (play-track 10))))))
    
    (gtk-container-border-width win gui-box-spacing)
    (gtk-container-add win vbox)
    (gtk-box-pack-start vbox (make-status-group) #f #f)
    (gtk-box-pack-start
     vbox
     (let ((hbox (gtk-hbox-new #f gui-box-spacing)))
       (gtk-box-pack-start
        hbox
        (let ((vbox (gtk-vbox-new #f gui-box-spacing)))
          (gtk-box-pack-start vbox (make-command-group) #f #f)
          (make-locatorA-group #f)
          (let* ((frame-and-groupA (make-locatorA-group #f))
                 (frameA (car frame-and-groupA))
                 (frame-and-groupX
                  (make-locatorX-group (cadr frame-and-groupA)))
                 (frameX (car frame-and-groupX)))
            (gtk-box-pack-start vbox frameA #f #f)
            (gtk-box-pack-start vbox frameX #f #f))
          (gtk-box-pack-start vbox (make-locatorB-group) #f #f)
          vbox))

       (gtk-box-pack-start
        hbox
        (let ((vbox (gtk-vbox-new #f gui-box-spacing)))
          (gtk-box-pack-start vbox (make-track-table) #f #f)
          (gtk-box-pack-start vbox (make-gap-group) #f #f)
          (gtk-box-pack-start
           vbox
           (let ((clear-button (make-button "Clear Loop" do-clear-loop)))
             (add-hook! gap-learn-mode-changed-hook
                        (lambda (state)
                          (gtk-widget-set-sensitive
                           clear-button
                              (not state))))
             clear-button) #f #f)

          (gtk-box-pack-start
           vbox
           (let ((auto-button (gtk-toggle-button-new-with-label
                           "Auto set X and gap with B")))
             (gtk-toggle-button-set-state auto-button autoset-locatorX)
             (gtk-signal-connect
              auto-button
              "toggled"
              (lambda ()
                (set! autoset-locatorX
                      (gtk-toggle-button-active auto-button))))
             auto-button) #f #f)

          vbox) #f #f)
       (if cdloop-mixer-enabled
           (gtk-box-pack-start hbox (make-mixer-group) #f #f))
       hbox) #f #f)
    
    (gtk-box-pack-start vbox (make-bookmark-group) #t #t)
    (gtk-box-pack-start
     vbox
     (let ((hbox (gtk-hbox-new #f gui-box-spacing)))
       (gtk-box-pack-start hbox (make-button "Quit" do-quit) #t #t)
       (gtk-box-pack-start hbox (make-button "About " show-about) #f #f)
       hbox) #f #f)
    win))

;;;}}}
;;;}}}
;;;{{{ main
(let ((options (getopt-long (program-arguments)
                            '((help (single-char #\h))
                              (version (single-char #\v))
                              (load (single-char #\l)
                                    (value #t))
                              (eval (single-char #\e)
                                    (value #t))
                              (device (single-char #\d)
                                      (value #t))
                              (no-mixer)
                              (bookmark (single-char #\b)
                                        (value #t))))))
  (cond
   ((assq 'help options)
    (display cdloop-about-text)
    (newline)
    (display cdloop-help-text)
    (quit))
   ((assq 'version options)
    (display (string-append "cdloop " cdloop-version))
    (newline)
    (quit))
   ((assq 'load options)
    (set! cdloop-config-file (cdr (assq 'load options)))))

  (set! cdloop-config-file (expand-filename cdloop-config-file))
  (if (access? cdloop-config-file R_OK)
      (load cdloop-config-file))

  ;; commandline overrides load file
  (if (assq 'device options)
      (set! cdloop-device (cdr (assq 'device options))))
  (if (assq 'no-mixer options)
      (set! cdloop-mixer-enabled #f))
  (if (assq 'bookmark options)
      (set! cdloop-bookmark-file (cdr (assq 'bookmark options))))
  (if (assq 'eval options)
      (eval (call-with-input-string (cdr (assq 'eval options)) read)))
  )

(set! cdloop-bookmark-file (expand-filename cdloop-bookmark-file))
(if (access? cdloop-bookmark-file R_OK)
    (load cdloop-bookmark-file))

(add-hook! toc-changed-hook
           (lambda ()
             (set-current-bookmark #f)
             (run-hook bookmarks-changed-hook)))
(add-hook! locatorA-changed-hook set-loop-if-valid)
(add-hook! locatorB-changed-hook set-loop-if-valid)
(add-hook! locatorX-changed-hook set-loop-if-valid)
(add-hook! gap-changed-hook set-loop-if-valid)
(add-hook! loop-start-changed-hook set-loop-if-valid)
(add-hook! status-changed-hook (lambda (status)
                                 (if (eq? status 'STATUS_NO_DISC)
                                     (do-clear-loop))))
(cdplayer-init cdloop-device)
(if (eq? (cdplayer-status) 'STATUS_NO_DEVICE)
    (begin
      (display (string-append
"\nCouldn't access " cdloop-device ".\n"
"You must have read permissions for " cdloop-device "\n\n"
))
      (quit)))

(cdplayer-mixer-init)
(if (not (cdplayer-mixer-valid))
    (set! cdloop-mixer-enabled #f))

(use-modules (gtk gtk)) ;; these need not be loaded if we only want to
(use-modules (gtk gdk)) ;; see the help text, so we load them late

(let ((toplevel (make-gui)))
  (gtk-widget-show-all toplevel)
  (set-loop-start 'A)
  (reset-mixer)
  (set! timeout-handle
        (gtk-timeout-add
         cdloop-gui-update-interval-milliseconds
         (let ((old-status #f)
               (old-pos #f))
           (lambda ()
             (let ((status (cdplayer-status)))
               (cond ((not (equal? status old-status))
                      (run-hook status-changed-hook status)
                      (set! old-status status))))
             (let ((pos (cdplayer-pos)))
               (cond ((not (equal? pos old-pos))
                      (run-hook pos-changed-hook pos)
                      (set! old-pos pos))))
             (cond
              ((cdplayer-toc-changed)
               (run-hook toc-changed-hook)))
             #t))))
  (gtk-standalone-main toplevel))
;;;}}}

;; Local variables:
;; folded-file: t
;; end:
